from kubernetes import client, config
import logging
from pathlib import Path
import uuid
from tenacity import retry, stop_after_attempt, wait_fixed


def generate_k8_pods(given_custom_images, given_namespace_name, num_pods, list_filename, file_dir):
    # NOTE: Assuming image and cluster already exist

    config.load_kube_config()
    apps_v1 = client.CoreV1Api()

    # *************
    # * Namespace *
    # *************
    # --- Creation of namespace
    logging.info('---- Namespace Creation Started ----')

    try:
        namespace_metadata = client.V1ObjectMeta(name=given_namespace_name)
        apps_v1.create_namespace(
            client.V1Namespace(metadata=namespace_metadata)
        )
    except Exception:
        # return immediately if given namespace already exists OR environment setup is incorrect (ex. doesn't have running cluster with image)
        logging.info('---- Creation Failed.... Terminating ----')
        raise Exception

    # --- Namespace creation
    logging.info('---- Namespace Successfully Added ----')

    # **************
    # * Config Map *
    # **************
    logging.info('---- Config Map Creation Started ----')
    
    # --- Config Map Meta Data
    config_map_meta_data = client.V1ObjectMeta(
        name='ms-config',
        namespace=given_namespace_name
    )

    # --- Put the content of the test files as key-value pair to Config Map
    config_map_data = {}  # dictionary to store data of configMap
    try:
        for test_file_path in list_filename:
            # Relative Path
            file = open(test_file_path, 'r')
            file_content = file.read()
            
            file_name = Path(test_file_path).name
            
            config_map_data.update({file_name:file_content})
            file.close()
    except Exception:
        logging.info('---- Test File Not Found, Deleting Allocated Resources... ----')
        delete_k8_deployment(given_namespace_name)
        raise Exception

    config_map_body = client.V1ConfigMap(
        api_version='v1',
        kind='ConfigMap',
        metadata=config_map_meta_data,
        data=config_map_data
    )

    apps_v1.create_namespaced_config_map(
        namespace=given_namespace_name,
        body=config_map_body
    )

    logging.info("---- Config Map Successfully Created ----")

    # ***********************
    # * Adding a Deployment *
    # ***********************

    if num_pods < len(given_custom_images):
        logging.info("---- Not enough workers. Terminating... ----")
        raise ValueError('Not Enough Workers Given')

    num_pods_allocated = num_pods // len(given_custom_images)
    for idx, custom_image in enumerate(given_custom_images):

        if idx == len(given_custom_images) - 1:
            num_pods_allocated = num_pods
        
        num_pods = num_pods - num_pods_allocated

        logging.info('---- Deployment Creation Started ----')
        
        # --- METADATA
        delpoyment_body_metadata = client.V1ObjectMeta(
            name=f'ms-deployment-{uuid.uuid4()}', 
            namespace=given_namespace_name, 
            labels={'app':'ms'}
        )

        # --- INNER SPEC-CONTAINER
        # Volume Mount List
        volume_mount_list = []
        volume_mount1 = client.V1VolumeMount(
            name='config-volume',
            mount_path=f'/code/{file_dir}'
        )
        volume_mount_list.append(volume_mount1)

        # Container inside the inner spec of template
        containers = []
        container1 = client.V1Container(
            name='ms-container', 
            image=custom_image, 
            image_pull_policy='IfNotPresent',
            volume_mounts=volume_mount_list
        )
        containers.append(container1)

        # Volume Config Map Info
        items_list = []
        for test_file_path in list_filename:
            # Get name of file, give to item as key-path pair (to include files in configMap)
            file_name = Path(test_file_path).name
            item = client.V1KeyToPath(
                key=file_name,
                path=file_name
            )
            items_list.append(item)

        volumes_config_map_info = client.V1ConfigMapVolumeSource(
            name='ms-config',
            items=items_list
        )

        # Volume that goes into inner spec
        volume_list = []
        volume1 = client.V1Volume(
            name='config-volume',
            config_map=volumes_config_map_info
        )
        volume_list.append(volume1)

        spec_template_spec = client.V1PodSpec(
            containers=containers,
            volumes=volume_list
        )

        # --- SPEC-SELECTOR-MATCH LABEL
        spec_selector = client.V1LabelSelector(
            match_labels={'app':'ms'}
        )

        # --- TEMPLATE
        template_metadata = client.V1ObjectMeta(
            labels={'app':'ms'}
        )
        spec_template = client.V1PodTemplateSpec(
            metadata=template_metadata, 
            spec=spec_template_spec
        )

        # --- SPEC
        deployment_body_spec = client.V1DeploymentSpec(
            replicas=num_pods_allocated, 
            selector=spec_selector, 
            template=spec_template
        )

        # --- CONFIG & CLIENT
        core_v1 = client.AppsV1Api()

        # --- ENTIRE YAML
        body = client.V1Deployment(
            api_version='apps/v1', 
            kind='Deployment', 
            metadata=delpoyment_body_metadata, 
            spec=deployment_body_spec
        )

        # --- CREATE DEPLOYMENT
        core_v1.create_namespaced_deployment(
            namespace=given_namespace_name, 
            body=body
        )

        # NOTE: wait until all the pods are in ready state    
        try:
            retry_check_pod_status(apps_v1, given_namespace_name)
        except:
            logging.info('---- Environment Failed to Initialize. Deleting Allocated Resources... ----')
            delete_k8_deployment(given_namespace_name)
            raise Exception

        logging.info('---- Deployment Successfully Created ----')


def delete_k8_deployment(given_namespace_name):
    # --- Deleting namespace at the end deletes all the related pods
    apps_v1 = client.CoreV1Api()
    apps_v1.delete_namespace(name=given_namespace_name)
    logging.info('---- Namespace Successfully Deleted ----')


@retry(
    wait=wait_fixed(5),
    stop=stop_after_attempt(10)
)
def retry_check_pod_status(apps_v1, given_namespace_name):
    # NOTE: Check if pods are in ready state
    logging.info('---- Waiting For Pods to be Prepared... ----')
    cnt = 0
    list_namespace_pod = apps_v1.list_namespaced_pod(given_namespace_name)
    if list_namespace_pod.items:
        for element in list_namespace_pod.items:
            pod_status = element.status.container_statuses[0]
            if pod_status.ready is True and pod_status.started is True:
                cnt += 1

        if cnt == len(list_namespace_pod.items):
            return

    raise Exception


from .Constants import NAMESPACE_NAME, CUSTOM_IMAGE_NAME
from .k8_client import generate_k8_pods, delete_k8_deployment
import logging
import pytest
import xdist
from kubernetes import client
from kubernetes import config as kubectl_config
from kubernetes.stream import stream
import uuid
import subprocess
from execnet import XSpec
from pathlib import Path
import socket
from time import sleep


def pytest_addoption(parser):
    parser.addoption(
        "--namespace", action="store", default=NAMESPACE_NAME, help="Defines the namespace of pods"
    )
    parser.addoption(
        "--custom_image", action="store", default=CUSTOM_IMAGE_NAME, help="Defines the name of the custom image"
    )


def pytest_xdist_setupnodes(config: pytest.Config, specs: list[XSpec]):

    if config.known_args_namespace.tx[0] != 'pod':
        return
    
    global process_list     # list of subprocesses responsible for port-forwarding
    process_list = []
    global ws_list          # list of streams for each pod' 
    ws_list = []
    
    # **********
    # * Set up *
    # **********
    global selected_namespace
    selected_namespace = config.option.namespace

    custom_image_list = config.option.custom_image.split(',')
    list_of_test_files = config.option.file_or_dir      # List of pytest files to run provided through the terminal

    # Find the path of packages and test files
    pkg_dir = Path(__file__).parent
    test_dir = Path(list_of_test_files[0]).parent

    # Move in all the necessary files from the package
    list_of_test_files.append(f"{pkg_dir}/ms_socketserver.py") # Move server.py that will run from the pod
    list_of_test_files.append(f"{pkg_dir}/conftest.py")

    if selected_namespace == NAMESPACE_NAME:
        selected_namespace = selected_namespace + '-' + str(uuid.uuid4())
  
    num_pods = len(specs)
    generate_k8_pods(
        given_custom_images=custom_image_list, 
        given_namespace_name=selected_namespace, 
        num_pods=num_pods, 
        list_filename=list_of_test_files, 
        file_dir=test_dir
    ) 

    # ***************************
    # * Communication with Pods *
    # ***************************

    kubectl_config.load_kube_config()
    api_instance = client.CoreV1Api()
    
    # NOTE: Go through all existing pods with the same namespace
    list_namespace_pod = api_instance.list_namespaced_pod(selected_namespace)
    for idx, np in enumerate(list_namespace_pod.items):
        k8_pod_name = np.metadata.name

        # Create a stream
        exec_command = ['/bin/sh']
        try:
            ws = stream(
                api_instance.connect_get_namespaced_pod_exec,
                k8_pod_name, selected_namespace,
                command=exec_command,
                stderr=True, stdin=True,
                stdout=True, tty=False,
                _preload_content=False
            )
            ws_list.append(ws)
        except:
            logging.info('---- Stream cannot be generated. Make sure that the pod is running ----')
            delete_k8_deployment(selected_namespace)
            raise Exception

        # Remote address - localhost : TCP port pair
        peer_pair = ws.sock.sock.getpeername()
        assigned_port = peer_pair[1]

        # NOTE: select the "available TCP port" using temporary sockets
        s = socket.socket()
        s.bind(('', 0))
        available_port = s.getsockname()[1]
        s.close()

        # NOTE: Running port-forward in background, give os.setsid to fully delete the thread at the end
        try:
            process = subprocess.Popen(
                ["kubectl", "port-forward", f"{k8_pod_name}", "--namespace", f"{selected_namespace}", f"{available_port}:{assigned_port}"],
                start_new_session=True
            )
            process_list.append(process)
        except:
            logging.info('---- Subprocess cannot be created. Check if the port is already occupied ----')
            delete_k8_deployment(selected_namespace)
            raise Exception

        logging.info("Subprocess running for port-forwarding")

        # NOTE: Run the server.py from each pod to listen to localhost
        commands = [
            f"python /code/{test_dir}/ms_socketserver.py :{assigned_port}"
        ]
        try:
            while ws.is_open():
                ws.update(timeout=1)
                if commands:
                    cm = commands.pop(0)
                    ws.write_stdin(cm + "\n")
                else:
                    break
        except:
            logging.info("Issue related to stream happened.. Terminating")
            delete_k8_deployment(selected_namespace)
            raise Exception

        logging.info("Server file listening to port")

        specs[idx].socket = f'127.0.0.1:{available_port}'
        specs[idx].popen = False

        # NOTE: Bypass xdist checking whether the directory exists inside the k8 pod
        config.pluginmanager.get_plugin('dsession').nodemanager.roots.append(test_dir)
        config.pluginmanager.get_plugin('dsession').nodemanager._rsynced_specs.add((specs[idx], test_dir))

    # NOTE: Give extra time for threads to port-forward and run server.py from each pod
    logging.info('---- Give Extra Time for Threads to Run Tasks... ----')
    sleep(2)


def pytest_sessionfinish(session):

    # NOTE: If controller, clean the allocated resources
    if xdist.is_xdist_controller(session):

        if session.config.known_args_namespace.tx[0] != 'pod':
            return

        # Kill process and child processes
        for process in process_list:
            process.kill()

        # Close the stream
        for ws in ws_list:
            ws.close()

        delete_k8_deployment(selected_namespace)
